#!/bin/sh

usage() {
	echo "Usage: $0 <[-BbfCa]> [-h]
Options:
  -B | --bar                   Fill signal bar with delay
  -b | --boot                  Show booting animation
  -f | --flash                 Show flashing animation
  -C | --clean                 Turn off all leds and exit
  -a | --auto                  Return to autonomous operation
  -c | --config                Configure leds using config file
  -D | --debug                 Enable debugging output
  -d | --daemon                Run as daemon
  -F | --force                 Force start/ignores 'ubusd' system and other 'ledman' instances
  -h | --help                  Print this message"
}

gpio_dir() {
	[ -n "$2" ] && echo "$2" > /sys/class/gpio/$1/direction
}

led_manual() {
	echo none > /sys/class/leds/$1/trigger
	[ -n "$2" ] && echo "$2" > /sys/class/leds/$1/brightness
}


led_timer() {
	echo timer > /sys/class/leds/$1/trigger
	[ -n "$2" ] && echo "$2" > /sys/class/leds/$1/delay_on
	[ -n "$2" ] && echo "$3" > /sys/class/leds/$1/delay_off
}

led_switch() {
	echo switch0 > /sys/class/leds/$1/trigger
	[ -n "$2" ] && echo $2 > /sys/class/leds/$1/port_mask
	[ -n "$3" ] && echo $3 > /sys/class/leds/$1/speed_mask
	[ -n "$4" ] && echo "$4" > /sys/class/leds/$1/mode
}

#############
#  Generic  #
#############

generic_bar() {
	"$1"_reload
	led_timer power_led 1000 1000
}

generic_boot() {
	"$1"_reload
	led_timer power_led 500 500
}

generic_flash() {
	"$1"_reload
	led_timer power_led 125 125
}

generic_auto() {
	"$1"_reload
}

generic_daemon() {
	"$1"_reload
}

generic_clean() {
	"$1"_clean
}

############
#  TAP100  #
############

tap100_reload() {
	led_switch eth0_led "0x10"
	led_manual gp_led 0
	led_manual power_led 1
}

tap100_clean() {
	led_manual eth0_led 0
	led_manual gp_led 0
	led_manual power_led 0
}

############
#  TAP200  #
############

# TODO: update and test with new TAP200 hardware
tap200_reload() {
	led_switch eth0_yellow_led "0x10"
	led_switch eth0_green_led "0x10"
	led_manual power_led 1
}

tap200_clean() {
	led_manual eth0_yellow_led 0
	led_manual eth0_green_led 0
	led_manual power_led 0
}

############
#  TSW202  #
############

tsw202_set_gpio_dir() {
	#killall other ledman instances
	ps | grep -E "ledman \-" | awk -v me=$$ 'me != $1 {print $1}' | xargs kill 2> /dev/null

	for gp in $(ls /sys/class/leds | grep port); do
		led_manual "$gp" 0
	done

	for gp in $(ls /sys/class/gpio | grep gpio4); do
		gpio_dir "$gp" "$1"
	done
}

tsw202_reload() {
	tsw202_set_gpio_dir "in"
}

tsw202_bar_bg() {
	for idx in $(seq 1 2 8); do
		led_manual "yellow:eth_port_${idx}" 1
		led_manual "green:eth_port_${idx}" 1
		led_manual "yellow:eth_port_$((idx+1))" 1
		led_manual "green:eth_port_$((idx+1))" 1
		sleep 2
	done

	led_manual "yellow:sfp_port_1" 1
	led_manual "yellow:sfp_port_2" 1
}

tsw202_bar() {
	tsw202_set_gpio_dir "out"

	tsw202_bar_bg &
}

tsw202_boot() {
	tsw202_set_gpio_dir "out"
	for gp in $(ls /sys/class/leds | grep port); do
		led_timer "$gp" 500 500
	done
}

tsw202_flash_bg() {
	while true; do
		# forwards
		for idx in $(seq 1 2 8); do
			led_manual "yellow:eth_port_${idx}" 1
			led_manual "yellow:eth_port_$((idx+1))" 1
			usleep 125000
			led_manual "yellow:eth_port_${idx}" 0
			led_manual "yellow:eth_port_$((idx+1))" 0
			led_manual "green:eth_port_${idx}" 1
			led_manual "green:eth_port_$((idx+1))" 1
			usleep 125000
			led_manual "green:eth_port_${idx}" 0
			led_manual "green:eth_port_$((idx+1))" 0
		done

		led_manual "yellow:sfp_port_2" 1
		usleep 125000
		led_manual "yellow:sfp_port_2" 0
		led_manual "yellow:sfp_port_1" 1
		usleep 125000
		led_manual "yellow:sfp_port_1" 0
		led_manual "yellow:sfp_port_2" 1
		usleep 125000
		led_manual "yellow:sfp_port_2" 0

		# backwards
		for idx in $(seq 7 -2 3); do
			led_manual "green:eth_port_${idx}" 1
			led_manual "green:eth_port_$((idx+1))" 1
			usleep 125000
			led_manual "yellow:eth_port_${idx}" 1
			led_manual "yellow:eth_port_$((idx+1))" 1
			led_manual "green:eth_port_${idx}" 0
			led_manual "green:eth_port_$((idx+1))" 0
			usleep 125000
			led_manual "yellow:eth_port_${idx}" 0
			led_manual "yellow:eth_port_$((idx+1))" 0
		done

		led_manual "green:eth_port_1" 1
		led_manual "green:eth_port_2" 1
		usleep 125000
		led_manual "green:eth_port_1" 0
		led_manual "green:eth_port_2" 0
	done 
}

tsw202_flash() {
	tsw202_set_gpio_dir "out"

	tsw202_flash_bg &
}

tsw202_clean() {
	tsw202_set_gpio_dir "out"
	for gp in $(ls /sys/class/leds | grep port); do
		led_manual "$gp" 0
	done
}

tsw202_auto() {
	tsw202_reload
}

tsw202_daemon() {
	tsw202_reload
}

##########
#  MAIN  #
##########

while getopts "BbfFCacdDh-:" opt; do
	case $opt in
		-)
			case ${OPTARG} in
				help)
					usage
					exit 0
					;;

				bar)
					ACTION="bar"
					;;
				boot)
					ACTION="boot"
					;;
				flash)
					ACTION="flash"
					;;
				clean)
					ACTION="clean"
					;;
				auto)
					ACTION="auto"
					;;
				daemon)
					[ -z "$ACTION" ] && ACTION="daemon"
					;;

				config |\
				debug |\
				force)
					echo "Unsupported action: '--${OPTARG}'"
					;;
				*)
					echo "Option does not exist: '--${OPTARG}'"
					usage
					exit 1
					;;

				esac
				;;

		h)
			usage
			exit 0
			;;

		B)
			ACTION="bar"
			;;
		b)
			ACTION="boot"
			;;
		f)
			ACTION="flash"
			;;
		C)
			ACTION="clean"
			;;
		a)
			ACTION="auto"
			;;
		d)
			[ -z "$ACTION" ] && ACTION="daemon"
			;;

		c|D|F)
			echo "Unsupported action: '-${opt}'"
			;;
		*)
			echo "Option does not exist: '-${opt}'"
			usage
			exit 1
			;;

	esac
done
shift $((OPTIND-1))

if [ "$(type -P mnf_info)" ]; then
	dev_name="$(mnf_info -n)"
else # Running in RAMDISK i.e. (no mnf_info binnary)
	dev_name="$(dd if=/dev/mtd1 bs=1 skip=$((0x10)) count=12)"
fi

[ -z "$ACTION" ] && {
	usage
	exit 1
}

[ "$(uci get system.ledman.enabled 2>/dev/null)" = "0" ]  && ACTION="clean"

case "$dev_name" in
	TAP100*)
		eval "generic_${ACTION}" "tap100"
		;;
	TAP200*)
		eval "generic_${ACTION}" "tap200"
		;;
	TSW202*|TSW212*)
		eval "tsw202_${ACTION}"
		;;
	*)
		echo "Unsupported device"
		exit 1
		;;
esac
